Attribute VB_Name = "modEvaluate"
'********************************************************************************
'* Author:         Adam Wehmann
'* Built Using:    Visual Basic 6.0 Learning Edition
'* Created:        8/05/2002
'* Last Modified:  8/05/2002
'********************************************************************************

Option Explicit

'*******************************************************************
'* Routine:        CountChar
'* Description:    Counts the amount of times one string occurs in another.
'* Parameters:     strText (String): String to search in
'*                 strChar (String): String to search for
'* Returns:        (Long) The number of times one string occurs in another.
'*******************************************************************
Private Function CountChar(strText As String, strChar As String) As Long
Dim lngPosition As Long
    'Make sure a valid string has been inputted.
    If Len(strChar) = 0 Then
        CountChar = 0
        Exit Function
    End If
    'Find the first occurence of strChar
    lngPosition = InStr(1, strText, strChar, vbTextCompare)
    If lngPosition <> 0 Then
        'Count characters.
        CountChar = CountChar + 1
        Do
            lngPosition = InStr(lngPosition + 1, strText, strChar, vbTextCompare)
            If lngPosition <> 0 Then
                CountChar = CountChar + 1
            End If
        Loop Until lngPosition = 0
    End If
End Function

'*******************************************************************
'* Routine:        Evaluate_Expression
'* Description:    Evaluates a mathematical expression consisting of
'*                 the following characters: 01234567890+-/*().
'* Parameters:     strExpression (String): Mathematical expression.
'* Returns:        (String) The number your expression evaluates to.
'*                 On failure the string may contain one of the following:
'*                      "error" - parenthesis don't match up.
'*******************************************************************
Public Function Evaluate_Expression(strExpression As String) As String
Dim strText As String, strLastText As String, regExpression As New RegExp
Dim mtchMatch As Match, mtchMatchCol As MatchCollection, lngCount As Long
Dim lngOpenParenthesis As Long, lngClosedParenthesis As Long
    strText = strExpression
    'Removes spaces from string.
    strText = Replace(strExpression, " ", "", 1, -1, vbTextCompare)
    'Count the amount of open and closing parenthesis in string.
    lngOpenParenthesis = CountChar(strText, "(")
    lngClosedParenthesis = CountChar(strText, ")")
    If lngOpenParenthesis <> lngClosedParenthesis Then
        'The number of open and closed parenthesis contained in the string are
        'unequal which means somewhere there is a mismatch. (5*(5-2)
        Evaluate_Expression = "error"
        Exit Function
    End If
    'Change #( to #*( since the parenthesis will be broken down into one
    'number each progressively starting from the innermost and going to
    'the outermost.  Otherwise the parenthesis would be broken down into
    'a number which would be placed on the side of the other number... ##
    Do
        strLastText = strText
        strText = RegularExpression_Find(strText, "([\d\.\)])+\(", 4)
    Loop While strText <> strLastText
    'Change )# to )*# since the parenthesis will be broken down into one
    'number each progressively starting from the innermost and going to
    'the outermost.  Otherwise the parenthesis would be broken down into
    'a number which would be placed on the side of the other number... ##
    Do
        strLastText = strText
        strText = RegularExpression_Find(strText, "\)([\d\.])+", 5)
    Loop While strText <> strLastText
    'Get rid of minus-a-negative's
    strText = Replace(strText, "--", "+", 1, -1, vbTextCompare)
    'Change minus signs (but not not negative signs) to plus negative.
    'A number minus a number is the same as a number plus negative a number.
    '1 - 1 = 1 + -1 which allows us to deal with negative numbers.
    With regExpression
        .Global = True
        .IgnoreCase = True
        .MultiLine = True
        .Pattern = "[^\+\(]-"
        Set mtchMatchCol = .Execute(strText)
        lngCount = 1
        For Each mtchMatch In mtchMatchCol
            If mtchMatch.Value <> "" Then
                strText = MidReplace(strText, mtchMatch.FirstIndex + lngCount, mtchMatch.Length, Mid(mtchMatch.Value, 1, 1) & "+-")
                lngCount = lngCount + 1
            End If
        Next
    End With
    'Evalutate the expression
    Evaluate_Expression = Evaluate_Expression2(strText)
End Function

'*******************************************************************
'* Routine:        Evaluate_Expression2
'* Description:    Evaluates a mathematical expression consisting of
'*                 the following characters: 01234567890+-/*().
'* Parameters:     strExpression (String): Mathematical expression.
'* Returns:        (String) The number your expression evaluates to.
'*******************************************************************
Private Function Evaluate_Expression2(strExpression As String) As String
Dim strText As String, lngOpenParenthesis As Long, strAnswer As String
Dim lngClosedParenthesis As Long, strLastText As String
Dim lngParenthesis As Long
    strText = strExpression
    'The following loop extracts parenthesis starting from the innermost and traveling
    'to the outermost fowards.  For example the expression ((3-1)+(1+1))((8/2)+(-2+6))
    'would be evaluated like this:
    '   3-1     and     1+1     so      (2+2)((8/2)+(-2+6))
    '   8/2     and     -2+6    so      (2+2)(4+4)
    '   2+2     and     4+4     so      32
    'After each parenthesis is extracted it is evaluated by calling this function
    'with the parenthesis' value. For example the expression ((-7+9)+(1+1))+4/(4-2)
    'would be evaluated like this:
    '   (2+(1+1))+4/(4-2)
    '   (2+2)+4/(4-2)
    '   4+4/(4-2)
    '   4+4/2
    '   4+2
    '   6
    lngOpenParenthesis = 0
    lngClosedParenthesis = 0
    Do
        lngOpenParenthesis = InStr(lngOpenParenthesis + 1, strText, "(", vbTextCompare)
        lngClosedParenthesis = InStr(lngOpenParenthesis + 1, strText, ")", vbTextCompare)
        'The following code handles multiple parenthesis pairs such as in this expression:
        '   ((3-1)+(1+1))((8/2)+(-2+6))
        lngParenthesis = InStr(lngOpenParenthesis + 1, strText, "(", vbTextCompare)
        If lngParenthesis < lngClosedParenthesis Then
            Do
                If InStr(lngParenthesis + 1, strText, "(", vbTextCompare) < lngClosedParenthesis And InStr(lngParenthesis + 1, strText, "(", vbTextCompare) <> 0 Then
                    lngParenthesis = InStr(lngParenthesis + 1, strText, "(", vbTextCompare)
                Else
                    Exit Do
                End If
            Loop While lngParenthesis < lngClosedParenthesis
            If lngParenthesis <> 0 Then
                lngOpenParenthesis = lngParenthesis
            End If
        End If
        If lngOpenParenthesis <> 0 And lngClosedParenthesis <> 0 Then
            'Evaluate our extracted parenthesis
            strAnswer = Evaluate_Expression2(Mid(strText, lngOpenParenthesis + 1, lngClosedParenthesis - (lngOpenParenthesis + 1)))
            strText = MidReplace(strText, lngOpenParenthesis, lngClosedParenthesis - lngOpenParenthesis + 1, strAnswer)
        End If
    Loop While lngOpenParenthesis <> 0 Or lngClosedParenthesis <> 0
    'Evaluate our extracted expression that should contain no parenthesis
    Do
        strLastText = strText
        'Evaluate division and multiplication.
        strText = Replace_DivisionAndMultiplication(strText)
    Loop While strText <> strLastText
    Do
        strLastText = strText
        'Evaluate addition.
        strText = Replace_Addition(strText)
    Loop While strText <> strLastText
    Evaluate_Expression2 = strText
End Function

'*******************************************************************
'* Routine:        MidReplace
'* Description:    Inserts one string into another.
'* Parameters:     strString (String): string to replace in.
'*                 lngStart (Long): starting position of text to replace.
'*                 lngLength (Long): length of text to replace.
'*                 strReplace (String): the replacement text.
'* Returns:        (String) Replaced string.
'*******************************************************************
Private Function MidReplace(strString As String, lngStart As Long, lngLength As Long, strReplace As String) As String
Dim strFirst As String, strSecond As String
    strFirst = Mid(strString, 1, lngStart - 1)
    strSecond = Mid(strString, lngStart + lngLength, Len(strString) - (lngStart + lngLength) + 1)
    MidReplace = strFirst & strReplace & strSecond
End Function

'*******************************************************************
'* Routine:        Replace_Addition
'* Description:    Evaluates all addition expressions.
'* Parameters:     strExpression (String): expression to evaluate.
'* Returns:        (String) evaluated string.
'*******************************************************************
Private Function Replace_Addition(strExpression As String) As String
    'Subtraction would go here... but we converted subtraction to plus negative's so
    'we no longer have need for it.
    strExpression = RegularExpression_Find(strExpression, "([\d\.-])+\+([\d\.-])+", 2)
    Replace_Addition = strExpression
End Function

'*******************************************************************
'* Routine:        Replace_DivisionAndMultiplication
'* Description:    Evaluates all division and multiplication expressions.
'* Parameters:     strExpression (String): expression to evaluate.
'* Returns:        (String) evaluated string.
'*******************************************************************
Private Function Replace_DivisionAndMultiplication(strExpression As String) As String
    strExpression = RegularExpression_Find(strExpression, "([\d\.-])+/([\d\.-])+", 0)
    strExpression = RegularExpression_Find(strExpression, "([\d\.-])+\*([\d\.-])+", 1)
    Replace_DivisionAndMultiplication = strExpression
End Function

'*******************************************************************
'* Routine:        Regular_Expression_Find
'* Description:    Evaluates different types of expressions.
'* Parameters:     strExpression (String): expression to evaluate.
'*                 strPattern (String): regular expression pattern
'*                      to use in searching.
'*                 lngOperation (Long): type of expression (see the
'*                      select case statement in code for numbers
'*                      and their corresponding operations.)
'* Returns:        (String) evaluated string.
'*******************************************************************
Private Function RegularExpression_Find(strExpression As String, strPattern As String, lngOperation As Long)
Dim regExpression As New RegExp, mtchMatch As Match, mtchMatchCol As MatchCollection
Dim strArray() As String, sngAnswer As Single
    With regExpression
        .IgnoreCase = True
        .Global = True
        .Pattern = strPattern
        Set mtchMatchCol = .Execute(strExpression)
        For Each mtchMatch In mtchMatchCol
            If mtchMatch.Value <> "" Then
                Select Case lngOperation
                Case 0 'Divides two numbers
                    strArray = Split(mtchMatch.Value, "/", -1, vbTextCompare)
                    sngAnswer = CSng(strArray(0)) / CSng(strArray(1))
                    strExpression = Replace(strExpression, mtchMatch.Value, CStr(sngAnswer), 1, 1)
                Case 1 'Multiplies two numbers
                    strArray = Split(mtchMatch.Value, "*", -1, vbTextCompare)
                    sngAnswer = CSng(strArray(0)) * CSng(strArray(1))
                    strExpression = Replace(strExpression, mtchMatch.Value, CStr(sngAnswer), 1, 1)
                Case 2 'Adds two numbers
                    strArray = Split(mtchMatch.Value, "+", -1, vbTextCompare)
                    sngAnswer = CSng(strArray(0)) + CSng(strArray(1))
                    strExpression = Replace(strExpression, mtchMatch.Value, CStr(sngAnswer), 1, 1)
                Case 3 'Subtracts two numbers
                    strArray = Split(mtchMatch.Value, "-", -1, vbTextCompare)
                    sngAnswer = CSng(strArray(0)) - CSng(strArray(1))
                    strExpression = Replace(strExpression, mtchMatch.Value, CStr(sngAnswer), 1, 1)
                Case 4 '#(
                    strExpression = Replace(strExpression, mtchMatch.Value, Mid(mtchMatch.Value, 1, Len(mtchMatch.Value) - 1) & "*(", 1, 1, vbTextCompare)
                Case 5 ')#
                    strExpression = Replace(strExpression, mtchMatch.Value, ")*" & Mid(mtchMatch.Value, 2, Len(mtchMatch.Value) - 1), 1, 1, vbTextCompare)
                End Select
            End If
        Next
    End With
    RegularExpression_Find = strExpression
End Function
